import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import yarl

from signalr_async.connection import ConnectionBase
from signalr_async.exceptions import HandshakeError

from .messages import HubInvocableMessage, HubMessage, PingMessage
from .protocols import JsonProtocol, MessagePackProtocol


class SignalRCoreConnection(ConnectionBase[HubMessage, HubInvocableMessage]):
    negotiate_version: str = "1"

    def __init__(
        self,
        base_url: str,
        hub_name: str,
        protocol: Optional[Union[JsonProtocol, MessagePackProtocol]] = None,
        extra_params: Optional[Dict[str, str]] = None,
        extra_headers: Optional[Dict[str, str]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(base_url, [hub_name], extra_params, extra_headers, logger)
        self._url = self._base_url / hub_name
        self.protocol = protocol or JsonProtocol()
        self.handshake_protocol = JsonProtocol()
        self._ping_message = self.protocol.write(PingMessage())

    def _common_params(self) -> Dict[str, str]:
        params = self._extra_params.copy()
        if self.connection_token:
            params["id"] = self.connection_token
        return params

    async def _send_command(
        self, command: str, params: Optional[Dict[str, str]] = None
    ) -> Any:
        if self._session is None:
            raise RuntimeError("Connection is not started")
        async with self._session.post(
            self._url / command,
            params=params or self._common_params(),
            headers=self._extra_headers,
        ) as resp:
            resp.raise_for_status()
            return await resp.json()

    async def _negotiate(self) -> None:
        negotiation_response = await self._send_command(
            "negotiate", params={"negotiateVersion": self.negotiate_version}
        )
        self.logger.debug(f"Negotiation response: {negotiation_response}")
        self.connection_id = negotiation_response["connectionId"]
        self.connection_token = negotiation_response.get(
            "connectionToken", self.connection_id
        )

    def _generate_connect_path(self) -> yarl.URL:
        if self._url.scheme == "http":
            scheme = "ws"
        elif self._url.scheme == "https":
            scheme = "wss"
        else:
            raise ValueError(f"Unknown scheme {self._url.scheme}")
        return self._url.with_scheme(scheme) % self._common_params()

    async def _initialize_connection(self) -> None:
        await self._send_raw(self.protocol.handshake_params, False)
        # TODO: timeout
        message = await self._receive_raw()
        handshake_result = list(
            self.handshake_protocol.decode(
                message.decode() if isinstance(message, bytes) else message
            )
        )[0]
        self.logger.debug(f"Handshake: {handshake_result}")
        if "error" in handshake_result:
            raise HandshakeError(handshake_result["error"])

    def _clear_connection_data(self) -> None:
        self.connection_id = None
        self.connection_token = None

    def _read_message(self, data: Union[str, bytes]) -> List[HubMessage]:
        # TODO: use self.protocol.is_binary instead of isinstance
        if isinstance(data, str) and isinstance(self.protocol, JsonProtocol):
            return list(self.protocol.parse(data))
        elif isinstance(data, bytes) and isinstance(self.protocol, MessagePackProtocol):
            return list(self.protocol.parse(data))
        else:
            raise RuntimeError(
                f"Protocol type is not respected. Got {type(data)}, expected {self.protocol.transfer_format}"
            )

    def _write_message(self, message: HubMessage) -> Tuple[Union[str, bytes], bool]:
        return self.protocol.write(message), self.protocol.is_binary

    async def ping(self) -> None:
        return await self._send_raw(self._ping_message, self.protocol.is_binary)



from typing import Any, Sequence

from signalr_async.hub import HubBase

from .messages import HubInvocation


class SignalRHub(HubBase[HubInvocation]):
    def _create_invocation_message(
        self, invocation_id: str, method: str, args: Sequence[Any]
    ) -> HubInvocation:
        return HubInvocation(
            invocation_id=invocation_id,
            hub=self.name,
            target=method,
            arguments=args or tuple(),
            state=None,
        )



from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence, Union

from signalr_async.messages import InvocationBase


@dataclass
class HubInvocation(InvocationBase):
    hub: str
    target: str
    arguments: Sequence[Any]
    state: Optional[Dict[str, Any]]

    @classmethod
    def from_raw_message(cls, raw_message: Dict[str, Any]) -> "HubInvocation":
        return cls(
            invocation_id=raw_message.get("I"),
            hub=raw_message["H"],
            target=raw_message["M"],
            arguments=raw_message["A"],
            state=raw_message.get("S"),
        )

    def to_raw_message(self) -> Dict[str, Any]:
        result = {
            "H": self.hub,
            "M": self.target,
            "A": self.arguments,
        }
        if self.invocation_id:
            result["I"] = self.invocation_id
        return result


@dataclass
class HubResult:
    invocation_id: str
    result: Optional[Any]
    error: Optional[str]
    error_data: Optional[Any]
    is_hub_exception: Optional[bool]
    progress_update: Optional[Any]
    state: Optional[Any]

    @classmethod
    def from_raw_message(cls, raw_message: Dict[str, Any]) -> "HubResult":
        return cls(
            invocation_id=raw_message["I"],
            result=raw_message.get("R"),
            error=raw_message.get("E"),
            error_data=raw_message.get("D"),
            is_hub_exception=raw_message.get("H"),
            progress_update=raw_message.get("P"),
            state=raw_message.get("S"),
        )


HubMessage = Union[HubInvocation, HubResult]



import asyncio
import logging
from abc import abstractmethod
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from typing_extensions import TypeGuard

from .invoke_manager import InvokeManager
from .messages import InvocationBase

SyncFunctionType = Callable[..., None]
AsyncFunctionType = Callable[..., Awaitable[None]]
CallbackType = Union[SyncFunctionType, AsyncFunctionType]
DecoratorType = Callable[[CallbackType], CallbackType]

I = TypeVar("I", bound=InvocationBase)


class HubBase(Generic[I]):
    def __init__(self, name: Optional[str] = None):
        self.name: str = name or type(self).__name__
        self._invoke_manager: Optional[InvokeManager[I]] = None
        self._logger: Optional[logging.Logger] = None
        self._callbacks: Dict[str, CallbackType] = {}
        for name in dir(self):
            if name.startswith("on_"):
                event_name = name[len("on_") :]
                self.on(event_name, getattr(self, name))

    def on(
        self, name: Optional[str] = None, callback: Optional[CallbackType] = None
    ) -> Union[CallbackType, DecoratorType]:
        def set_callback_decorator(callback: CallbackType) -> CallbackType:
            self._callbacks[name or callback.__name__] = callback
            return callback

        return set_callback_decorator(callback) if callback else set_callback_decorator

    def _set_invoke_manager(self, invoke_manager: InvokeManager[I]) -> "HubBase[I]":
        self._invoke_manager = invoke_manager
        return self

    def _set_logger(self, logger: logging.Logger) -> "HubBase[I]":
        self._logger = logger
        return self

    async def _call(self, method_name: str, args: Sequence[Any]) -> None:
        callback = self._callbacks.get(method_name)
        if callback is not None:
            asyncio.create_task(self._async_callback(callback, args))
        elif self._logger is not None:
            self._logger.warning(
                f"Method {method_name} doesnt exist in hub {self.name}"
            )

    def _is_coroutine_function(
        self, callback: CallbackType
    ) -> TypeGuard[AsyncFunctionType]:
        return asyncio.iscoroutinefunction(callback)

    async def _async_callback(self, callback: CallbackType, args: Sequence[Any]) -> Any:
        if self._is_coroutine_function(callback):
            return await callback(*args)
        else:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, callback, *args)

    async def invoke(self, method: str, *args: Any) -> Any:
        if self._invoke_manager is None:
            raise RuntimeError(f"Hub {self.name} is not registered")
        invocation_id = self._invoke_manager.next_invocation_id()
        message = self._create_invocation_message(invocation_id, method, args)
        return await self._invoke_manager.invoke_and_wait_for_result(
            invocation_id, message
        )

    @abstractmethod
    def _create_invocation_message(
        self, invocation_id: str, method: str, args: Sequence[Any]
    ) -> I:
        """Create invocation message"""

    async def on_connect(self, connection_id: str) -> None:
        """Connect event"""

    async def on_disconnect(self) -> None:
        """Disconnect event"""



import asyncio
from typing import Any, Dict, Sequence

from signalr_async.client import SignalRClientBase

from .connection import SignalRConnection
from .hub import SignalRHub

# from .invoke_manager import SignalRInvokeManager
from .messages import HubInvocation, HubMessage, HubResult


class SignalRClient(SignalRClientBase[Sequence[SignalRHub], HubMessage, HubInvocation]):
    def build_connection(
        self,
        base_url: str,
        connection_options: Dict[str, Any],
    ) -> SignalRConnection:
        # TODO: Fix in base
        self._hub_dict = {h.name: h for h in self._hub}
        return SignalRConnection(
            base_url=base_url,
            hub_names=[hub.name for hub in self._hub],
            extra_params=connection_options.get("extra_params"),
            extra_headers=connection_options.get("extra_headers"),
            logger=self.logger,
        )

    async def _connection_event(self) -> None:
        if self._connection.connection_id:
            for hub in self._hub:
                asyncio.create_task(hub.on_connect(self._connection.connection_id))

    async def _disconnection_event(self) -> None:
        for hub in self._hub:
            asyncio.create_task(hub.on_disconnect())

    async def _process_message(self, message: HubMessage) -> None:
        if isinstance(message, HubResult):
            if message.progress_update:
                self.logger.error(
                    f"ProgressUpdate messages are not implemented: {message}"
                )
            elif message.result:
                self._invoke_manager.set_invocation_result(
                    message.invocation_id, message.result
                )
            elif message.error:
                self._invoke_manager.set_invocation_exception(
                    message.invocation_id, message.error
                )
            else:
                self.logger.error(f"Bad HubResult: {message}")
        elif isinstance(message, HubInvocation):
            if message.hub:
                await self._hub_dict[message.hub]._call(
                    message.target, message.arguments
                )



import json
from abc import ABC, abstractmethod
from typing import Any, Generator, Generic, TypeVar

from signalr_async.netcore.messages import HubMessage

T = TypeVar("T")


class ProtocolBase(ABC, Generic[T]):
    seperator: str = chr(0x1E)

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the protocol"""

    @property
    @abstractmethod
    def version(self) -> int:
        """Version of the protocol"""

    @property
    @abstractmethod
    def is_binary(self) -> bool:
        """Does the protocol use binary frames"""

    @property
    def transfer_format(self) -> str:
        return "Binary" if self.is_binary else "Text"

    @property
    def handshake_params(self) -> str:
        return (
            json.dumps({"protocol": self.name, "version": self.version})
            + self.seperator
        )

    @abstractmethod
    def decode(self, raw_messages: T) -> Generator[Any, None, None]:
        """Generator which decodes messages from connection to python objects"""

    @abstractmethod
    def parse(self, raw_messages: T) -> Generator[HubMessage, None, None]:
        """Parse raw message from connection to HubMessge using decode method"""

    @abstractmethod
    def encode(self, output: Any) -> T:
        """Encode python objects to transferable format"""

    @abstractmethod
    def write(self, message: HubMessage) -> T:
        """Transform HubMessage to transferable format"""



from typing import Any, Generator, Sequence, Tuple

import msgpack

from signalr_async.exceptions import InvalidMessage
from signalr_async.netcore.messages import (
    CancelInvocationMessage,
    CloseMessage,
    CompletionMessage,
    HubMessage,
    InvocationMessage,
    MessageTypes,
    PingMessage,
    StreamInvocationMessage,
    StreamItemMessage,
)

from .base import ProtocolBase


class MessagePackProtocol(ProtocolBase[bytes]):
    max_length_prefix_size: int = 5

    _error_result = 1
    _void_result = 2
    _non_void_result = 3

    @property
    def name(self) -> str:
        return "messagepack"

    @property
    def version(self) -> int:
        return 1

    @property
    def is_binary(self) -> bool:
        return True

    def _get_size(self, message: bytes, offset: int) -> Tuple[int, int]:
        total_readable_bytes = len(message) - offset
        num_bytes = 0
        size = 0
        while True:
            read_byte = message[offset + num_bytes]
            size = size | ((read_byte & 0x7F) << 7 * num_bytes)
            num_bytes += 1
            if num_bytes == self.max_length_prefix_size:
                if read_byte > 3:
                    raise InvalidMessage("Messages bigger than 2GB are not supported.")
                else:
                    return num_bytes, size
            elif (read_byte & 0x80) == 0:
                return num_bytes, size
            elif num_bytes >= total_readable_bytes:
                raise InvalidMessage("Cant read message size.")

    def decode(self, raw_messages: bytes) -> Generator[Sequence[Any], None, None]:
        offset = 0
        while offset < len(raw_messages):
            num_bytes, size = self._get_size(raw_messages, offset)
            if offset + num_bytes + size > len(raw_messages):
                raise InvalidMessage("Incomplete message.")
            yield msgpack.unpackb(
                raw_messages[offset + num_bytes : offset + num_bytes + size]
            )
            offset += num_bytes + size

    def parse(self, raw_messages: bytes) -> Generator[HubMessage, None, None]:
        for message in self.decode(raw_messages):
            message_type = message[0]
            if message_type == MessageTypes.INVOCATION:
                yield InvocationMessage(
                    headers=message[1],
                    invocation_id=message[2],
                    target=message[3],
                    arguments=message[4],
                    stream_ids=message[5],
                )
            elif message_type == MessageTypes.STREAM_ITEM:
                yield StreamItemMessage(
                    headers=message[1], invocation_id=message[2], item=message[3]
                )
            elif message_type == MessageTypes.COMPLETION:
                result_kind = message[3]
                yield CompletionMessage(
                    headers=message[1],
                    invocation_id=message[2],
                    error=message[4] if result_kind == self._error_result else None,
                    result=message[4] if result_kind == self._non_void_result else None,
                )
            elif message_type == MessageTypes.PING:
                yield PingMessage()
            elif message_type == MessageTypes.CLOSE:
                yield CloseMessage(
                    error=message[1],
                    allow_reconnect=message[2] if len(message) >= 3 else None,
                )
            else:
                raise RuntimeError("Unknown message")

    def encode(self, output: Sequence[Any]) -> bytes:
        encoded_output = msgpack.packb(output)
        size = len(encoded_output)
        length_prefix = b""
        while size > 0:
            size_part = size & 0x7F
            size >>= 7
            if size > 0:
                size_part |= 0x80
            length_prefix += size_part.to_bytes(1, "big")
        return length_prefix + encoded_output

    def write(self, message: HubMessage) -> bytes:
        output: Sequence[Any]
        if isinstance(message, InvocationMessage):
            output = [
                message.message_type,
                message.headers,
                message.invocation_id,
                message.target,
                message.arguments,
                message.stream_ids,
            ]
        elif isinstance(message, StreamInvocationMessage):
            output = [
                message.message_type,
                message.headers,
                message.invocation_id,
                message.target,
                message.arguments,
                message.stream_ids,
            ]
        elif isinstance(message, StreamItemMessage):
            output = [
                message.message_type,
                message.headers,
                message.invocation_id,
                message.item,
            ]
        elif isinstance(message, CompletionMessage):
            result_kind = (
                self._error_result
                if message.error
                else (self._non_void_result if message.result else self._void_result)
            )
            output = [
                message.message_type,
                message.headers,
                message.invocation_id,
                result_kind,
            ]
            output.append(result_kind)
            if result_kind != self._void_result:
                output.append(message.error or message.result)
        elif isinstance(message, PingMessage):
            output = [message.message_type]
        elif isinstance(message, CancelInvocationMessage):
            output = [
                message.message_type,
                message.headers,
                message.invocation_id,
            ]
        else:
            raise RuntimeError("Unknown message type")
        return self.encode(output)



import logging
import time
from abc import abstractmethod
from typing import Dict, Generic, List, Optional, Tuple, TypeVar, Union

import aiohttp
import yarl

from signalr_async.exceptions import ConnectionClosed, ConnectionInitializationError

T = TypeVar("T")
O = TypeVar("O")


class ConnectionBase(Generic[T, O]):
    def __init__(
        self,
        base_url: str,
        hub_names: Optional[List[str]] = None,
        extra_params: Optional[Dict[str, str]] = None,
        extra_headers: Optional[Dict[str, str]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self._base_url = yarl.URL(base_url)
        self._extra_params = extra_params or {}
        self._extra_headers = extra_headers or {}
        self.logger = logger or logging.getLogger(__name__)
        self._hub_names = hub_names or []
        self.last_message_received_time: Optional[float] = None
        self.last_message_sent_time: Optional[float] = None
        self._session: Optional[aiohttp.ClientSession] = None
        self._websocket: Optional[aiohttp.ClientWebSocketResponse] = None
        self.connection_id: Optional[str] = None
        self.connection_token: Optional[str] = None
        self.state = "disconnected"

    async def start(self) -> bool:
        self.logger.debug(f"Starting connection with {self.state} state")
        if self.state == "disconnected":
            try:
                self.state = "connecting"
                self._session = aiohttp.ClientSession()
                self.logger.debug("Negotiation started")
                await self._negotiate()
                connect_path = self._generate_connect_path()
                self.logger.debug(f"Connecting to {connect_path}")
                self._websocket = await self._session.ws_connect(
                    connect_path, headers=self._extra_headers
                )
                self.logger.debug("Initializing")
                await self._initialize_connection()
                self.state = "connected"
                return True
            except aiohttp.client_exceptions.ClientError as e:
                self.logger.debug("Starting connection failed")
                await self.stop()
                raise ConnectionInitializationError("Client cannot connect") from e
            except ConnectionInitializationError:
                self.logger.debug("initialize connection failed")
                await self.stop()
                raise
        return False

    @abstractmethod
    async def _negotiate(self) -> None:
        pass

    @abstractmethod
    def _generate_connect_path(self) -> yarl.URL:
        pass

    @abstractmethod
    async def _initialize_connection(self) -> None:
        pass

    async def stop(self) -> bool:
        self.logger.debug(f"Stopping connection with {self.state} state")
        if self.state not in ("disconnecting", "disconnected"):
            self.state = "disconnecting"
            self._clear_connection_data()
            if self._websocket is not None:
                await self._websocket.close()
                self._websocket = None
                self.logger.debug("Websocket closed")
            if self._session is not None:
                await self._session.close()
                self._session = None
                self.logger.debug("Session closed")
            self.last_message_received_time = None
            self.last_message_sent_time = None
            self.state = "disconnected"
            return True
        return False

    def _clear_connection_data(self) -> None:
        pass

    async def _receive_raw(self, timeout: Optional[float] = None) -> Union[str, bytes]:
        if self._websocket is None:
            raise ConnectionClosed() from None
        raw_ws_message = await self._websocket.receive(timeout=timeout)
        self.logger.debug(f"Raw message received: {raw_ws_message}")
        if raw_ws_message.type in (
            aiohttp.WSMsgType.CLOSE,
            aiohttp.WSMsgType.CLOSING,
            aiohttp.WSMsgType.CLOSED,
        ):
            raise ConnectionClosed(raw_ws_message.data, raw_ws_message.extra) from None
        # elif raw_ws_message.type not in (
        #     aiohttp.WSMsgType.TEXT,
        #     aiohttp.WSMsgType.BINARY,
        # ):
        #     raise Exception("Unknown websocket message type") from None
        self.last_message_received_time = time.time()
        return raw_ws_message.data  # type: ignore

    async def receive(self, timeout: Optional[float] = None) -> List[T]:
        return self._read_message(await self._receive_raw(timeout=timeout))

    async def _send_raw(
        self, message_content: Union[str, bytes], is_binary: bool
    ) -> None:
        self.logger.debug(f"Sending: {message_content=}, {is_binary=}")
        if self._websocket is None:
            raise ConnectionClosed() from None
        try:
            await self._websocket._writer.send(
                message_content, is_binary, compress=None
            )
            self.last_message_sent_time = time.time()
        except ConnectionResetError as e:
            raise ConnectionClosed() from e

    async def send(self, message: O) -> None:
        return await self._send_raw(*self._write_message(message))

    @abstractmethod
    def _read_message(self, data: Union[str, bytes]) -> List[T]:
        pass

    @abstractmethod
    def _write_message(self, message: O) -> Tuple[Union[str, bytes], bool]:
        pass

    @abstractmethod
    async def ping(self) -> None:
        pass



class AsyncSignalRException(Exception):
    pass


class AsyncSignalRConnectionError(AsyncSignalRException, ConnectionError):
    pass


class ConnectionInitializationError(AsyncSignalRConnectionError):
    pass


class HandshakeError(ConnectionInitializationError):
    pass


class IncompatibleServerError(AsyncSignalRConnectionError):
    pass


class ConnectionClosed(AsyncSignalRConnectionError):
    pass


class ServerInvocationException(AsyncSignalRException):
    pass


class MessageError(AsyncSignalRException):
    pass


class InvalidMessage(MessageError):
    pass



from typing import Any, AsyncGenerator, Sequence

from signalr_async.hub import HubBase

from .messages import (
    CancelInvocationMessage,
    HubInvocableMessage,
    InvocationMessage,
    StreamInvocationMessage,
)


class SignalRCoreHub(HubBase[HubInvocableMessage]):
    def _create_invocation_message(
        self, invocation_id: str, method: str, args: Sequence[Any]
    ) -> InvocationMessage:
        return InvocationMessage(
            invocation_id=invocation_id,
            target=method,
            arguments=args or tuple(),
            headers={},
            stream_ids=tuple(),
        )

    async def stream(self, method: str, *args: Any) -> AsyncGenerator[Any, None]:
        if self._invoke_manager is None:
            raise RuntimeError(f"Hub {self.name} is not registered")
        invocation_id = self._invoke_manager.next_invocation_id()
        message = StreamInvocationMessage(
            invocation_id=invocation_id,
            target=method,
            arguments=args,
            headers={},
            stream_ids=tuple(),
        )
        async for result in self._invoke_manager.invoke_and_wait_for_stream(
            invocation_id, message
        ):
            try:
                yield result
            except GeneratorExit:
                if self._logger:
                    self._logger.debug(f"Closing stream {invocation_id} with {method=}")
                await self._invoke_manager.invoke(
                    CancelInvocationMessage(invocation_id=invocation_id, headers={})
                )
                break



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['signalr_async',
 'signalr_async.net',
 'signalr_async.netcore',
 'signalr_async.netcore.protocols']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.0,<4.0.0', 'msgpack>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'signalr-async',
    'version': '2.2.1',
    'description': 'Python SignalR async client',
    'long_description': '# SignalR-Async\n\n<p align="center">\n<a href="https://app.travis-ci.com/sam-mosleh/signalr-async" target="_blank">\n    <img src="https://app.travis-ci.com/sam-mosleh/signalr-async.svg?branch=master" alt="Test">\n</a>\n\n<a href="https://codecov.io/gh/sam-mosleh/signalr-async">\n  <img src="https://codecov.io/gh/sam-mosleh/signalr-async/branch/master/graph/badge.svg?token=JYBKXSFAX6"/>\n</a>\n\n<a href="https://pypi.org/project/signalr-async/" target="_blank">\n    <img src="https://img.shields.io/pypi/v/signalr-async" alt="Package version">\n</a>\n<a href="https://pypi.org/project/signalr-async/" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/signalr-async.svg" alt="Supported Python versions">\n</a>\n</p>\n\nSignalR-Async is a python client for ASP.NET & ASP.NET Core SignalR, ready for building bidirectional communication.\n\n## Installation\n\n```bash\npip install signalr-async\n```\n\n## Example\n\n### Create it\n\n* Create a file `main.py` with:\n\n```Python\nimport asyncio\nfrom signalr_async.netcore import Hub, Client\nfrom signalr_async.netcore.protocols import MessagePackProtocol\n\n\nclass MyHub(Hub):\n    async def on_connect(self, connection_id: str) -> None:\n        """Will be awaited after connection established"""\n\n    async def on_disconnect(self) -> None:\n        """Will be awaited after client disconnection"""\n\n    def on_event_one(self, x: bool, y: str) -> None:\n        """Invoked by server synchronously on (event_one)"""\n\n    async def on_event_two(self, x: bool, y: str) -> None:\n        """Invoked by server asynchronously on (event_two)"""\n\n    async def get_something(self) -> bool:\n        """Invoke (method) on server"""\n        return await self.invoke("method", "arg1", 2)\n\n\nhub = MyHub("my-hub")\n\n\n@hub.on("event_three")\nasync def three(z: int) -> None:\n    pass\n\n\n@hub.on\nasync def event_four(z: int) -> None:\n    pass\n\n\nasync def multi_event(z: int) -> None:\n    pass\n\n\nfor i in range(10):\n    hub.on(f"event_{i}", multi_event)\n\n\nasync def main():\n    token = "mytoken"\n    async with Client(\n        "https://localhost:9000",\n        hub,\n        connection_options={\n            "extra_headers": {"Authorization": f"Bearer {token}"},\n            "protocol": MessagePackProtocol(),\n        },\n    ) as client:\n        return await hub.get_something()\n\n\nasyncio.run(main())\n```\n\n## Resources\n\nSee the [SignalR Documentation](https://docs.microsoft.com/aspnet/core/signalr) at docs.microsoft.com for documentation on the latest release.\n',
    'author': 'Sam Mosleh',
    'author_email': 'sam.mosleh@ut.ac.ir',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sam-mosleh/signalr-async',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
from dataclasses import asdict
from typing import Any, Dict, Generator, Tuple

from signalr_async.netcore.messages import (
    HubMessage,
    MessageTypes,
    message_type_to_class,
)

from .base import ProtocolBase


class JsonProtocol(ProtocolBase[str]):
    aliases = {
        "invocationId": "invocation_id",
        "streamIds": "stream_ids",
        "allowReconnect": "allow_reconnect",
    }
    aliases_reversed = {v: k for k, v in aliases.items()}

    @property
    def name(self) -> str:
        return "json"

    @property
    def version(self) -> int:
        return 1

    @property
    def is_binary(self) -> bool:
        return False

    def decode(self, raw_messages: str) -> Generator[Dict[str, Any], None, None]:
        for msg in raw_messages.split(self.seperator):
            if msg:
                yield json.loads(msg)

    def _sanitize_raw_message_dict(
        self, message: Dict[str, Any]
    ) -> Tuple[MessageTypes, Dict[str, Any]]:
        result = {self.aliases.get(key, key): val for key, val in message.items()}
        message_type = result.pop("type")
        if (
            message_type not in (MessageTypes.PING, MessageTypes.CLOSE)
            and "headers" not in result
        ):
            result["headers"] = {}
        if (
            message_type in (MessageTypes.INVOCATION, MessageTypes.STREAM_INVOCATION)
            and "stream_ids" not in result
        ):
            result["stream_ids"] = []
        if message_type == MessageTypes.INVOCATION:
            result["invocation_id"] = result.get("invocation_id")
        elif message_type == MessageTypes.COMPLETION:
            result["result"] = result.get("result")
            result["error"] = result.get("error")
        elif message_type == MessageTypes.CLOSE:
            result["error"] = result.get("error")
            result["allow_reconnect"] = result.get("allow_reconnect")
        return MessageTypes(message_type), result

    def parse(self, raw_messages: str) -> Generator[HubMessage, None, None]:
        for message in self.decode(raw_messages):
            message_type, sanitized = self._sanitize_raw_message_dict(message)
            yield message_type_to_class[message_type](**sanitized)  # type: ignore

    def encode(self, output: Dict[str, Any]) -> str:
        return json.dumps(output) + self.seperator

    def write(self, message: HubMessage) -> str:
        output = {
            self.aliases_reversed.get(key, key): val
            for key, val in asdict(message).items()
            if val is not None
        }
        output["type"] = message.message_type
        return self.encode(output)



import asyncio
from typing import Any, Dict

from signalr_async.client import SignalRClientBase

from .connection import SignalRCoreConnection
from .hub import SignalRCoreHub

# from .invoke_manager import SignalRCoreInvokeManager
from .messages import (
    CancelInvocationMessage,
    CloseMessage,
    CompletionMessage,
    HubInvocableMessage,
    HubMessage,
    InvocationMessage,
    PingMessage,
    StreamInvocationMessage,
    StreamItemMessage,
)


class SignalRCoreClient(
    SignalRClientBase[SignalRCoreHub, HubMessage, HubInvocableMessage]
):
    def build_connection(
        self,
        base_url: str,
        connection_options: Dict[str, Any],
    ) -> SignalRCoreConnection:
        return SignalRCoreConnection(
            base_url=base_url,
            hub_name=self._hub.name,
            protocol=connection_options.get("protocol"),
            extra_params=connection_options.get("extra_params"),
            extra_headers=connection_options.get("extra_headers"),
            logger=self.logger,
        )

    async def _connection_event(self) -> None:
        if self._connection.connection_id:
            asyncio.create_task(self._hub.on_connect(self._connection.connection_id))

    async def _disconnection_event(self) -> None:
        asyncio.create_task(self._hub.on_disconnect())

    async def _process_message(self, message: HubMessage) -> None:
        if isinstance(message, InvocationMessage):
            await self._hub._call(message.target, message.arguments)
        elif isinstance(message, StreamItemMessage):
            self._invoke_manager.add_invocation_result(
                message.invocation_id, message.item
            )
        elif isinstance(message, CompletionMessage):
            if message.error:
                self._invoke_manager.set_invocation_exception(
                    message.invocation_id, message.error
                )
            else:
                self._invoke_manager.set_invocation_result(
                    message.invocation_id, message.result
                )
        elif isinstance(message, StreamInvocationMessage):
            self.logger.debug(f"Stream callback {message}")
        elif isinstance(message, CancelInvocationMessage):
            self.logger.debug(f"Cancel {message}")
        elif isinstance(message, PingMessage):
            self.logger.debug("[PING]")
        elif isinstance(message, CloseMessage):
            if message.error:
                self.logger.error(f"Server closed with reason: {message.error}")
            if message.allow_reconnect:
                await self._stop_connection()
            else:
                await asyncio.shield(self.stop())
        else:
            raise RuntimeError("Unknown message type")



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Dict, Optional, Sequence, Type, Union

from signalr_async.messages import InvocationBase


class MessageTypes(IntEnum):
    INVOCATION = 1
    STREAM_ITEM = 2
    COMPLETION = 3
    STREAM_INVOCATION = 4
    CANCEL_INVOCATION = 5
    PING = 6
    CLOSE = 7


class HubMessageBase(ABC):
    @property
    @abstractmethod
    def message_type(self) -> MessageTypes:
        """Message type of SignalRCore messages"""


@dataclass
class HubInvocableMessageBase(InvocationBase):
    invocation_id: str
    headers: Dict[str, Any]


@dataclass
class InvocationMessage(HubMessageBase, InvocationBase):
    target: str
    arguments: Sequence[Any]
    headers: Dict[str, Any]
    stream_ids: Sequence[str]

    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.INVOCATION


@dataclass
class StreamItemMessage(HubMessageBase, HubInvocableMessageBase):
    item: Dict[str, Any]

    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.STREAM_ITEM


@dataclass
class CompletionMessage(HubMessageBase, HubInvocableMessageBase):
    error: Optional[str]
    result: Optional[Dict[str, Any]]

    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.COMPLETION


@dataclass
class StreamInvocationMessage(HubMessageBase, HubInvocableMessageBase):
    target: str
    arguments: Sequence[Any]
    stream_ids: Sequence[Any]

    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.STREAM_INVOCATION


@dataclass
class CancelInvocationMessage(HubMessageBase, HubInvocableMessageBase):
    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.CANCEL_INVOCATION


@dataclass
class PingMessage(HubMessageBase):
    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.PING


@dataclass
class CloseMessage(HubMessageBase):
    error: Optional[str]
    allow_reconnect: Optional[bool]

    @property
    def message_type(self) -> MessageTypes:
        return MessageTypes.CLOSE


HubMessage = Union[
    InvocationMessage,
    StreamItemMessage,
    CompletionMessage,
    StreamInvocationMessage,
    CancelInvocationMessage,
    PingMessage,
    CloseMessage,
]

HubInvocableMessage = Union[
    InvocationMessage,
    StreamItemMessage,
    CompletionMessage,
    StreamInvocationMessage,
    CancelInvocationMessage,
]

message_type_to_class: Dict[MessageTypes, Type[HubMessage]] = {
    MessageTypes.INVOCATION: InvocationMessage,
    MessageTypes.STREAM_ITEM: StreamItemMessage,
    MessageTypes.COMPLETION: CompletionMessage,
    MessageTypes.STREAM_INVOCATION: StreamInvocationMessage,
    MessageTypes.CANCEL_INVOCATION: CancelInvocationMessage,
    MessageTypes.PING: PingMessage,
    MessageTypes.CLOSE: CloseMessage,
}



import asyncio
import logging
import time
from abc import ABC, abstractmethod
from types import TracebackType
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
)

from .connection import ConnectionBase
from .exceptions import ConnectionClosed, ConnectionInitializationError
from .hub import HubBase
from .invoke_manager import InvokeManager
from .messages import InvocationBase

# Type of Invoke message which is in queue
I = TypeVar("I", bound=InvocationBase)
# Type of hub
# H = TypeVar("H")
H = TypeVar("H", bound=Union[HubBase[Any], Sequence[HubBase[Any]]])
# Type of all Messages received
R = TypeVar("R")


class SignalRClientBase(ABC, Generic[H, R, I]):
    def __init__(
        self,
        base_url: str,
        hub: H,
        timeout: float = 30,
        keepalive_interval: float = 15,
        reconnect_policy: bool = True,
        logger: Optional[logging.Logger] = None,
        connection_options: Optional[Dict[str, Any]] = None,
    ):
        self._hub = hub
        self.logger = logger or logging.getLogger(__name__)
        self.reconnect_policy = reconnect_policy
        self.timeout = timeout
        self.keepalive_interval = keepalive_interval
        self._all_tasks: Set["asyncio.Task[None]"] = set()
        self._producer_queue: "asyncio.Queue[I]" = asyncio.Queue()
        self._invoke_manager = InvokeManager(self._producer_queue)
        self._connection = self.build_connection(base_url, connection_options or {})
        if isinstance(self._hub, Sequence):
            for h in self._hub:
                h._set_invoke_manager(self._invoke_manager)._set_logger(self.logger)
        else:
            self._hub._set_invoke_manager(self._invoke_manager)._set_logger(self.logger)

    @abstractmethod
    def build_connection(
        self, base_url: str, connection_options: Dict[str, Any]
    ) -> ConnectionBase[R, I]:
        """Build new connection"""

    async def __aenter__(self) -> "SignalRClientBase[H, R, I]":
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.stop()

    async def start(self) -> None:
        if await self._start_connection():
            self._all_tasks.update(
                {
                    asyncio.create_task(self._consumer()),
                    asyncio.create_task(self._producer()),
                    asyncio.create_task(self._timeout_handler()),
                    asyncio.create_task(self._keepalive_handler()),
                }
            )
            # Context switch for tasks
            await asyncio.sleep(0)
            self.logger.debug("Tasks created")
        self.logger.debug("Client started successfully")

    async def stop(self) -> None:
        tasks: List["asyncio.Task[None]"] = []
        if self._all_tasks:
            tasks.extend(self._all_tasks)
            self._all_tasks.clear()
            for task in tasks:
                task.cancel()
        await self._stop_connection()
        if tasks:
            gather = await asyncio.gather(*tasks)
            self.logger.debug(f"{gather=}")

    async def _start_connection(self) -> bool:
        if await self._connection.start():
            await self._connection_event()
            return True
        return False

    async def _stop_connection(self) -> bool:
        if await self._connection.stop():
            await self._disconnection_event()
            return True
        return False

    @abstractmethod
    async def _connection_event(self) -> None:
        """Inform hub or hubs that client is connected"""

    @abstractmethod
    async def _disconnection_event(self) -> None:
        """Inform hub or hubs that client is disconnected"""

    async def _producer(self) -> None:
        while True:
            try:
                message = await self._producer_queue.get()
                if message.invocation_id:
                    try:
                        await self._connection.send(message)
                    except ConnectionClosed:
                        self.logger.error(
                            f"Message has not been sent because connection is closed"
                        )
                        self._invoke_manager.set_invocation_exception(
                            message.invocation_id, "Connection is closed"
                        )
            except asyncio.CancelledError:
                break

    async def _consumer(self) -> None:
        while True:
            try:
                try:
                    messages = await self._connection.receive()
                    for message in messages:
                        await self._process_message(message)
                except ConnectionClosed:
                    # Receive failure
                    self.logger.debug(f"Consumer state is {self._connection.state}")
                    if self.reconnect_policy:
                        if self._connection.state == "connected":
                            await self._stop_connection()
                        elif self._connection.state == "disconnected":
                            try:
                                await self._start_connection()
                            except ConnectionInitializationError as e:
                                self.logger.exception(e)
                    else:
                        await asyncio.shield(self.stop())
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                break

    @abstractmethod
    async def _process_message(self, message: R) -> None:
        """Process received messages"""

    async def _timeout_handler(self) -> None:
        timeout_steps = min(self.timeout / 10, 1)
        while True:
            try:
                await asyncio.sleep(timeout_steps)
                if self._connection.last_message_received_time is not None:
                    diff = time.time() - self._connection.last_message_received_time
                    if diff >= self.timeout:
                        self.logger.error("Connection timeout")
                        await self._stop_connection()
                    elif diff >= self.timeout * 0.8:
                        self.logger.warning("No message received in a long time")
            except asyncio.CancelledError:
                break

    async def _keepalive_handler(self) -> None:
        while True:
            try:
                if self._connection.last_message_sent_time is None:
                    await asyncio.sleep(1)
                else:
                    diff = time.time() - self._connection.last_message_sent_time
                    if diff >= self.keepalive_interval:
                        self.logger.debug("Connection keepalive")
                        try:
                            await self._connection.ping()
                        except ConnectionClosed:
                            self.logger.error("Connection keepalive failed to send")
                    else:
                        await asyncio.sleep(self.keepalive_interval - diff)
            except asyncio.CancelledError:
                break

    async def wait(self, timeout: Optional[float] = None) -> None:
        done, pending = await asyncio.wait(
            self._all_tasks, timeout=timeout, return_when=asyncio.FIRST_COMPLETED
        )
        for task in done:
            await task
        if self._all_tasks and done:
            raise RuntimeError("Tasks should be running")



import asyncio
from typing import Any, Dict, Generic, TypeVar

from .exceptions import ServerInvocationException
from .messages import InvocationBase

I = TypeVar("I", bound=InvocationBase)


class InvokeManager(Generic[I]):
    def __init__(self, queue: "asyncio.Queue[I]"):
        self._queue = queue
        self.invocation_events: Dict[str, asyncio.Event] = {}
        self.invocation_results: Dict[str, Any] = {}
        self.invocation_exceptions: Dict[str, ServerInvocationException] = {}
        self.total_invokes = 0

    def next_invocation_id(self) -> str:
        invocation_id = str(self.total_invokes)
        self.total_invokes += 1
        event = asyncio.Event()
        self.invocation_events[invocation_id] = event
        return invocation_id

    async def _wait_for(self, invocation_id: str) -> Any:
        if invocation_id not in self.invocation_events:
            raise RuntimeError(f"Invocation {invocation_id} not found")
        await self.invocation_events[invocation_id].wait()
        if invocation_id in self.invocation_exceptions:
            raise self.invocation_exceptions.pop(invocation_id)
        return self.invocation_results.pop(invocation_id)

    async def invoke(self, message: I) -> None:
        return await self._queue.put(message)

    async def invoke_and_wait_for_result(self, invocation_id: str, message: I) -> Any:
        await self.invoke(message)
        result = await self._wait_for(invocation_id)
        self.invocation_events.pop(invocation_id)
        return result

    async def invoke_and_wait_for_stream(self, invocation_id: str, message: I) -> Any:
        self.invocation_results[invocation_id] = []
        await self.invoke(message)
        while True:
            results = await self._wait_for(invocation_id)
            self.invocation_results[invocation_id] = []
            self.invocation_events[invocation_id].clear()
            for result in results:
                if isinstance(result, StopAsyncIteration):
                    return
                yield result

    def set_invocation_result(self, invocation_id: str, result: Any) -> None:
        if invocation_id not in self.invocation_events:
            raise RuntimeError(f"invocation event {invocation_id} not found")
        if isinstance(self.invocation_results.get(invocation_id), list):
            self.invocation_results[invocation_id].append(StopAsyncIteration())
        else:
            self.invocation_results[invocation_id] = result
        self.invocation_events[invocation_id].set()

    def add_invocation_result(self, invocation_id: str, result: Any) -> None:
        if invocation_id not in self.invocation_events:
            raise RuntimeError(f"invocation event {invocation_id} not found")
        if isinstance(self.invocation_results.get(invocation_id), list):
            self.invocation_results[invocation_id].append(result)
            self.invocation_events[invocation_id].set()
        else:
            raise RuntimeError(f"Invocation {invocation_id} is not a stream")

    def set_invocation_exception(self, invocation_id: str, error_msg: str) -> None:
        if invocation_id not in self.invocation_events:
            raise RuntimeError(f"invocation event {invocation_id} not found")
        self.invocation_exceptions[invocation_id] = ServerInvocationException(error_msg)
        self.invocation_events[invocation_id].set()



import json
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import yarl

from signalr_async.connection import ConnectionBase
from signalr_async.exceptions import (
    ConnectionInitializationError,
    IncompatibleServerError,
)

from .messages import HubInvocation, HubMessage, HubResult

# NegotiationResponse
# {
#     public string ConnectionId { get; set; }
#     public string ConnectionToken { get; set; }
#     public string Url { get; set; }
#     public string ProtocolVersion { get; set; }
#     public double DisconnectTimeout { get; set; }
#     public bool TryWebSockets { get; set; }
#     public double? KeepAliveTimeout { get; set; }
#     public double TransportConnectTimeout { get; set; }

#     // Protocol 2.0: Redirection and custom Negotiation Errors
#     public string RedirectUrl { get; set; }
#     public string AccessToken { get; set; }
#     public string Error { get; set; }
# }


class CommandEnum(str, Enum):
    NEGOTIATE = "negotiate"
    START = "start"
    CONNECT = "connect"
    RECONNECT = "reconnect"
    ABORT = "abort"


class SignalRConnection(ConnectionBase[HubMessage, HubInvocation]):
    client_protocol_version: str = "1.5"

    def __init__(
        self,
        base_url: str,
        hub_names: Optional[List[str]] = None,
        extra_params: Optional[Dict[str, str]] = None,
        extra_headers: Optional[Dict[str, str]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(base_url, hub_names, extra_params, extra_headers, logger)
        self.transport: Optional[str] = None
        self.message_id: Optional[str] = None
        self.groups_token: Optional[str] = None
        self.keepalive_timeout: Optional[int] = None
        self._ping_message = "{}"

    def _common_params(self) -> Dict[str, str]:
        connection_data = [{"name": hub_name} for hub_name in self._hub_names]
        params = dict(
            clientProtocol=self.client_protocol_version,
            connectionData=json.dumps(connection_data),
            **self._extra_params,
        )
        if self.transport is not None:
            params["transport"] = self.transport
        if self.connection_token is not None:
            params["connectionToken"] = self.connection_token
        return params

    def _receive_params(self) -> Dict[str, str]:
        params = self._common_params()
        if self.message_id is not None:
            params["messageId"] = self.message_id
        if self.groups_token is not None:
            params["groupsToken"] = self.groups_token
        return params

    async def _send_command(self, command: CommandEnum) -> Any:
        if self._session is None:
            raise RuntimeError("Connection is not started")
        async with self._session.get(
            self._base_url / command.value,
            params=self._common_params(),
            headers=self._extra_headers,
        ) as resp:
            resp.raise_for_status()
            return await resp.json()

    async def _negotiate(self) -> None:
        negotiation_response = await self._send_command(CommandEnum.NEGOTIATE)
        self.logger.debug(f"Negotiation response: {negotiation_response}")
        if "availableTransports" in negotiation_response:
            raise IncompatibleServerError("ASP.Net Core server detected")
        if negotiation_response["TryWebSockets"] is False:
            raise IncompatibleServerError(
                "Server does not support websocket as a transport."
            )
        self.transport = "webSockets"
        # TODO: raise IncompatibleServerError in case of bad ProtocolVersion
        self.connection_token = negotiation_response["ConnectionToken"]
        self.connection_id = negotiation_response["ConnectionId"]
        self.keepalive_timeout = negotiation_response.get("KeepAliveTimeout", 5)

    def _generate_receive_path(self, command: CommandEnum) -> yarl.URL:
        if self._base_url.scheme == "http":
            scheme = "ws"
        elif self._base_url.scheme == "https":
            scheme = "wss"
        else:
            raise ValueError(f"Unknown scheme {self._base_url.scheme}")
        return (
            self._base_url.with_scheme(scheme) / command.value % self._receive_params()
        )

    def _generate_connect_path(self) -> yarl.URL:
        return self._generate_receive_path(CommandEnum.CONNECT)

    async def _initialize_connection(self) -> None:
        # TODO: timeout
        # "C": str, S: int, M: List
        first_message = json.loads(await self._receive_raw())
        if "C" in first_message and first_message.get("S", 0) == 1:
            self.message_id = first_message["C"]
            start_response = await self._send_command(CommandEnum.START)
            if start_response["Response"] != "started":
                raise ConnectionInitializationError("Server is not started")
            await self.ping()
        else:
            raise ConnectionInitializationError("Server did not send the start message")

    def _read_message(self, data: Union[str, bytes]) -> List[HubMessage]:
        raw_message: Dict[str, Any] = json.loads(data)
        if "I" in raw_message:
            return [HubResult.from_raw_message(raw_message)]
        message_dict = {
            "message_id": raw_message.get("C"),
            "messages": raw_message.get("M"),
            "initialized": raw_message.get("S"),
            "should_reconnect": raw_message.get("T"),
            "long_poll_delay": raw_message.get("L"),
            "groups_token": raw_message.get("G"),
            "error": raw_message.get("E"),
        }
        if message_dict["error"] is not None:
            # TODO: Close connection with error
            self.logger.error(message_dict["error"])
        if message_dict["groups_token"] is not None:
            self.groups_token = message_dict["groups_token"]
        if message_dict["messages"] is not None:
            self.message_id = message_dict["message_id"]
            return [HubInvocation.from_raw_message(m) for m in message_dict["messages"]]
        elif raw_message:
            self.logger.error(f"Whats this: {raw_message}")
        return []

    def _write_message(self, message: HubInvocation) -> Tuple[Union[str, bytes], bool]:
        return json.dumps(message.to_raw_message()), False

    def _clear_connection_data(self) -> None:
        self.connection_id = None
        self.connection_token = None
        self.groups_token = None
        self.message_id = None

    async def ping(self) -> None:
        return await self._send_raw(self._ping_message, False)
